#include <stdio.h>
#include "stm32l0xx.h"
#include "usb.h"

// The current clock frequency
uint32_t SystemCoreClock=2097000;

// Counts milliseconds
volatile uint32_t systick_count=0;

// Interrupt handler
void SysTick_Handler(void)
{
    systick_count++;
}

// Delay some milliseconds
// Note that effective time may be up to 1ms shorter than requested.
void delay_ms(int ms)
{
    uint32_t start=systick_count;
    while (systick_count-start<ms);
}

// Change system clock to 32 MHz using internal 16 MHz R/C oscillator
void init_clock()
{
    // Because the debugger switches PLL on, we may need to switch
    // back to the HSI oscillator before we can configure the PLL

    // Enable HSI oscillator
    SET_BIT(RCC->CR, RCC_CR_HSION);

    // Wait until HSI oscillator is ready
    while(!READ_BIT(RCC->CR, RCC_CR_HSIRDY)) {}

    // Switch to HSI oscillator
    MODIFY_REG(RCC->CFGR, RCC_CFGR_SW, RCC_CFGR_SW_HSI);

    // Wait until the switch is done
    while ((RCC->CFGR & RCC_CFGR_SWS_Msk) != RCC_CFGR_SWS_HSI) {}

    // Disable the PLL
    CLEAR_BIT(RCC->CR, RCC_CR_PLLON);

    // Wait until the PLL is fully stopped
    while(READ_BIT(RCC->CR, RCC_CR_PLLRDY)) {}

    // Flash latency 1 wait state
    SET_BIT(FLASH->ACR, FLASH_ACR_LATENCY);

    // 32 MHz using the 16 MHz HSI oscillator multiply by 4 divide by 2
    WRITE_REG(RCC->CFGR, RCC_CFGR_PLLSRC_HSI + RCC_CFGR_PLLMUL4 + RCC_CFGR_PLLDIV2);

    // Switch the PLL on
    SET_BIT(RCC->CR, RCC_CR_PLLON);

    // Wait until the PLL is ready
    while(!READ_BIT(RCC->CR, RCC_CR_PLLRDY)) {}

    // Select the PLL as clock source
    MODIFY_REG(RCC->CFGR, RCC_CFGR_SW, RCC_CFGR_SW_PLL);

    // Update variable
    SystemCoreClock=32000000;

    // Switch the voltage reference for the HSI48 oscillator on
    SET_BIT(RCC->APB2ENR, RCC_APB2ENR_SYSCFGEN);
    SET_BIT(SYSCFG->CFGR3, SYSCFG_CFGR3_ENREF_HSI48);

    // Wait until the voltage reference is ready
    while(!READ_BIT(SYSCFG->CFGR3, SYSCFG_CFGR3_VREFINT_RDYF)) {}

    // Switch the HSI48 oscillator on
    SET_BIT(RCC->CRRCR, RCC_CRRCR_HSI48ON);

    // Wait until the HSI48 oscillator is ready
    while(!READ_BIT(RCC->CRRCR, RCC_CRRCR_HSI48RDY)) {}

    // Select the HSI48 oscillator as clock source for USB
    SET_BIT(RCC->CCIPR, RCC_CCIPR_HSI48SEL);
}

void init_io()
{
    // Enable Port A
    SET_BIT(RCC->IOPENR, RCC_IOPENR_GPIOAEN);

    // PA5 = Mode 01 Output
    MODIFY_REG(GPIOA->MODER, GPIO_MODER_MODE5, 0b01 << GPIO_MODER_MODE5_Pos);

    // Enable USB with pull-up
    SET_BIT(RCC->APB1ENR, RCC_APB1ENR_USBEN);
    SET_BIT(USB->BCDR, USB_BCDR_DPPU);
}

// Redirect standard output to the USB port, e.g. from puts() or printf()
int _write(int file, char *ptr, int len)
{
    return UsbSendBytes(ptr, len, 10);
}

int main()
{
    init_clock();
    init_io();
    UsbSetup();

    // Initialize system timer
    SysTick_Config(SystemCoreClock/1000);

    while (1)
    {
        // LED On (High)
        WRITE_REG(GPIOA->BSRR, GPIO_BSRR_BS_5);
        delay_ms(100);

        UsbSendStr("Hello World!\n",10);
        // Alternative: puts("Hello World!");

        // If a line has been received, then send an echo back
        if (UsbRxBufferContains('\n'))
        {
            char buf[256];
            UsbGetStr(buf,sizeof(buf),'\n',0);
            UsbSendStr(buf,10);
        }

        // LED Off (Low)
        WRITE_REG(GPIOA->BSRR, GPIO_BSRR_BR_5);
        delay_ms(900);
    }
}

